/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Matrix.h"
#include "Map.h"

using namespace FD;
using namespace std;

class MapMatrixValueOverwrite;

DECLARE_NODE(MapMatrixValueOverwrite)
/*Node
 *
 * @name MapMatrixValueOverwrite
 * @category HARK:MISC
 * @description Replaces specific cells of the Map Matrix values by a given value.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The input Matrix. Map<int,Matrix<int> >, Map<int,Matrix<float> > or Map<int,Matrix<complex<float> > >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The Matrix with the overwritten value. Map<int,Matrix<int> >, Map<int,Matrix<float> > or Map<int,Matrix<complex<float> > >.
 *
 * @parameter_name OVERWRITTEN_ROW_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the row where to start setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_ROW_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the row where to end setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_COL_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the column where to start setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_COL_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the column where to end setting the Matrix.
 *
 * @parameter_name OVERWRITE_VALUE_REAL
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Real part of setting value.
 *
 * @parameter_name OVERWRITE_VALUE_IMAG
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Imaginary part of setting value.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

class MapMatrixValueOverwrite : public BufferedNode {
	int inputID;
	int outputID;
	int rwMin;
	int rwMax;
	int clMin;
	int clMax;
	float fReal;
	float fImag;	
	bool debug;

public:
	MapMatrixValueOverwrite(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		rwMin = dereference_cast<int>(parameters.get("OVERWRITTEN_ROW_MIN"));
		rwMax = dereference_cast<int>(parameters.get("OVERWRITTEN_ROW_MAX"));
		clMin = dereference_cast<int>(parameters.get("OVERWRITTEN_COL_MIN"));
		clMax = dereference_cast<int>(parameters.get("OVERWRITTEN_COL_MAX"));
		fReal = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_REAL"));
		fImag = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_IMAG"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void checkSize(int &vMin, int &vMax, int vSize) {
		if (vMin < 0) vMin = 0;
		else if (vMin >= vSize) vMin = vSize - 1;
		if (vMax < 0) vMax = 0;
		else if (vMax >= vSize) vMax = vSize - 1;
		if (vMin > vMax) {
			cerr << "******* MapMatrixValueOverwrite: Minimun index is larger than maximum index." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MapMatrixValueOverwrite]", __FILE__, __LINE__));
		}
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;

		if (debug == true) {
			cout << "******* MapMatrixValueOverwrite debug point-Input check (count=" << count << ")" << endl;
 			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Matrix<int>))			  cout << "Matrix<int>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<float>))           cout << "Matrix<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		Map<int, ObjectRef>& output = *new Map<int, ObjectRef>;
		out[count] = &output;

		if (input_map.empty() == false) {
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Matrix<int>)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<int> &vcInput = object_cast<Matrix<int> >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					checkSize(rwMin, rwMax, rSize);
					checkSize(clMin, clMax, cSize);
					RCPtr<Matrix<int> > fbank(new Matrix<int>(rSize, cSize));
					/**fbank = vcInput;
					for (int i = rwMin; i <= rwMax; i++) {
						for (int j = clMin; j <= clMax; j++) {
							(*fbank)[i][j] = static_cast<int>(fReal);
						}
					}*/
					for (int i = 0; i < rSize; i++) {
						for (int j = 0; j < cSize; j++) {
							if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
								(*fbank)[i][j] = static_cast<int>(fReal);
							else
								(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<float>)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<float> &vcInput = object_cast<Matrix<float> >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					checkSize(rwMin, rwMax, rSize);
					checkSize(clMin, clMax, cSize);
					RCPtr<Matrix<float> > fbank(new Matrix<float>(rSize, cSize));
					/**fbank = vcInput;
					for (int i = rwMin; i <= rwMax; i++) {
						for (int j = clMin; j <= clMax; j++) {
							(*fbank)[i][j] = fReal;
						}
					}*/
					for (int i = 0; i < rSize; i++) {
						for (int j = 0; j < cSize; j++) {
							if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
								(*fbank)[i][j] = fReal;
							else
								(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<complex<float> > &vcInput = object_cast<Matrix<complex<float> > >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					checkSize(rwMin, rwMax, rSize);
					checkSize(clMin, clMax, cSize);
					RCPtr<Matrix<complex<float> > > fbank(new Matrix<complex<float> >(rSize, cSize));
					/**fbank = vcInput;
					for (int i = rwMin; i <= rwMax; i++) {
						for (int j = clMin; j <= clMax; j++) {
							(*fbank)[i][j] = complex<float>(fReal, fImag);
						}
					}*/
					for (int i = 0; i < rSize; i++) {
						for (int j = 0; j < cSize; j++) {
							if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
								(*fbank)[i][j] = complex<float>(fReal, fImag);
							else
								(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else {
				cerr << "******* MapMatrixValueOverwrite: Input type is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapMatrixValueOverwrite]", __FILE__, __LINE__));
			}
		}

        if (debug == true) cout << "******* MapMatrixValueOverwrite debug point-Finish" << endl;
	}

    NO_ORDER_NODE_SPEEDUP(MapMatrixValueOverwrite);
};
