/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"

using namespace FD;
using namespace std;

class MapToMap;

DECLARE_NODE(MapToMap)
/*Node
 *
 * @name MapToMap
 * @category HARK:MISC
 * @description Convert ObjectRef of Map<int, ObjectRef>. If the input is Map<int, Vector<float> >, the output is Map<int, Vector<complex<float> > >. If the input is Map<int, Vector<complex<float> > >, the output is Map<int, Vector<float> >. If the input is Map<int, Matrix<float> >, the output is Map<int, Matrix<complex<float> > >. If the input is Map<int, Matrixr<complex<float> > >, the output is Map<int, Matrix<float> >.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Vector<float> or Vector<complex<float> > or Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The key is source ID, and the value is Vector<float> or Vector<complex<float> > or Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name METHOD_COMPLEX_TO_FLOAT
 * @parameter_type string
 * @parameter_list magnitude:real:imaginary
 * @parameter_value magnitude
 * @parameter_description Method to convert float to complex. This is used only when INPUT is Map<int, Vector<float> > or Map<int, Matrix<float> >. [default: magnitude]
 *
 * @parameter_name METHOD_FLOAT_TO_COMPLEX
 * @parameter_type string
 * @parameter_list zero:hilbert
 * @parameter_value zero
 * @parameter_description Method to convert complex to float. This is used only when INPUT is Map<int, Vector<complex<float> > > or Map<int, Matrix<complex<float> > >. [default: zero]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int COMPLEX_ZERO = 0;
static const int COMPLEX_HILBERT = 1;

static const int SOURCE_MAGNITUDE = 0;
static const int SOURCE_REAL = 1;
static const int SOURCE_IMAGINARY = 2;

class MapToMap : public BufferedNode {
	int inputID;
	int outputID;
	int method;
	int complex_method;
	bool debug;

public:
	MapToMap(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_str = object_cast<String>(parameters.get("METHOD_COMPLEX_TO_FLOAT"));
		if (method_str == "magnitude")      method = SOURCE_MAGNITUDE;
		else if (method_str == "real")      method = SOURCE_REAL;
		else if (method_str == "imaginary") method = SOURCE_IMAGINARY;
		else {
			cerr << "******* MapToMap: Input error" << endl;
			throw new NodeException(NULL, "MapToMap: METHOD should be \"magnitude\" or \"real\" or \"imaginary\".", __FILE__, __LINE__);
		}
		string complex_str = object_cast<String>(parameters.get("METHOD_FLOAT_TO_COMPLEX"));
		if (complex_str == "zero")         complex_method = COMPLEX_ZERO;
		else if (complex_str == "hilbert") complex_method = COMPLEX_HILBERT;
		else {
			cerr << "******* MapToMap: Input error" << endl;
			throw new NodeException(NULL, "MapToMap: COMPLEX_METHOD should be \"zero\" or \"hilbert\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}
    
	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;

		if (debug == true) {
			cout << "******* MapToMap debug point-Input check (count=" << count << ")" << endl;

			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Vector<float>))           cout << "Vector<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<float>))           cout << "Matrix<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		ObjectRef orOutputTemp = ObjectRef(new Map<int, ObjectRef>);
		Map<int, ObjectRef> &mpOrOutput = object_cast<Map<int, ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;

		if (input_map.empty() == false) {
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Vector<float>)) {
				for (; it != input_map.end(); ++it) {
					const Vector<float> &vcFltInput = object_cast<Vector<float> >(it->second);
					ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(vcFltInput.size()));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
					if (complex_method == COMPLEX_HILBERT) {
						for (int col = 0; col < vcFltInput.size(); col++) {
							vcCmpOutput[col] = complex<float>(vcFltInput[col], abs(vcFltInput[col]));
						}
					}
					else { // complex_method == COMPLEX_ZERO
						for (int col = 0; col < vcFltInput.size(); col++) {
							vcCmpOutput[col] = complex<float>(vcFltInput[col], 0.0f);
						}
					}
					mpOrOutput.insert(make_pair(it->first, orOutputTemp));
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Vector<complex<float> > &vcCmpInput = object_cast<Vector<complex<float> > >(it->second);
					ObjectRef orOutputTemp = ObjectRef(new Vector<float>(vcCmpInput.size()));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
					if (method == SOURCE_REAL) {
						for (int col = 0; col < vcCmpInput.size(); col++) {
							vcFltOutput[col] = vcCmpInput[col].real();
						}
					}
					else if (method == SOURCE_IMAGINARY) {
						for (int col = 0; col < vcCmpInput.size(); col++) {
							vcFltOutput[col] = vcCmpInput[col].imag();
						}
					}
					else { // method == SOURCE_MAGNITUDE
						for (int col = 0; col < vcCmpInput.size(); col++) {
							vcFltOutput[col] = abs(vcCmpInput[col]);
						}
					}
					mpOrOutput.insert(make_pair(it->first, orOutputTemp));
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<float>)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<float> &mtFltInput = object_cast<Matrix<float> >(it->second);
					ObjectRef orOutputTemp = ObjectRef(new Matrix<complex<float> >(mtFltInput.nrows(), mtFltInput.ncols()));
					Matrix<complex<float> > &mtCmpOutput = object_cast<Matrix<complex<float> > >(orOutputTemp);
					if (complex_method == COMPLEX_HILBERT) {
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							for (int col = 0; col < mtFltInput.ncols(); col++) {
								mtCmpOutput(row, col) = complex<float>(mtFltInput(row, col), abs(mtFltInput(row, col)));
							}
						}
					}
					else { // complex_method == COMPLEX_ZERO
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							for (int col = 0; col < mtFltInput.ncols(); col++) {
								mtCmpOutput(row, col) = complex<float>(mtFltInput(row, col), 0.0f);
							}
						}
					}
					mpOrOutput.insert(make_pair(it->first, orOutputTemp));
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<complex<float> > &mtCmpInput = object_cast<Matrix<complex<float> > >(it->second);
					ObjectRef orOutputTemp = ObjectRef(new Matrix<float>(mtCmpInput.nrows(), mtCmpInput.ncols()));
					Matrix<float> &mtFltOutput = object_cast<Matrix<float> >(orOutputTemp);
					if (method == SOURCE_REAL) {
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							for (int col = 0; col < mtCmpInput.ncols(); col++) {
								mtFltOutput(row, col) = mtCmpInput(row, col).real();
							}
						}
					}
					else if (method == SOURCE_IMAGINARY) {
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							for (int col = 0; col < mtCmpInput.ncols(); col++) {
								mtFltOutput(row, col) = mtCmpInput(row, col).imag();
							}
						}
					}
					else { // method == SOURCE_MAGNITUDE
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							for (int col = 0; col < mtCmpInput.ncols(); col++) {
								mtFltOutput(row, col) = abs(mtCmpInput(row, col));
							}
						}
					}
					mpOrOutput.insert(make_pair(it->first, orOutputTemp));
				}
			}
			else {
				cerr << "******* MapToMap: Input type of ObjectRef is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapToMap]", __FILE__, __LINE__));
			}
		}

		if (debug == true) cout << "******* MapToMap debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(MapToMap);
};
