/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Map.h"

using namespace std;
using namespace FD;

class MapToVector;

DECLARE_NODE(MapToVector)
/*Node
 *
 * @name MapToVector
 * @category HARK:MISC
 * @description Convert Map<int, ObjectRef> to Vector<float> or Vector<complex<float> >. If the input is Map<int, Vecotr<float> >, the output is Vector<float>. If the input is Map<int, complex<float> > >, the output is Vector<complex<float> >.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Vector<float> or Vector<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Vector<float> or Vector<complex<float> >.
 *
 * @parameter_name METHOD
 * @parameter_type string
 * @parameter_list min_key:max_key:average:summation
 * @parameter_value min_key
 * @parameter_description Method of transform Map to Vector. [default: min_key]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int METHOD_MIN_KEY = 0;
static const int METHOD_MAX_KEY = 1;
static const int METHOD_AVERAGE = 2;
static const int METHOD_SUMMATION = 3;

class MapToVector : public BufferedNode {
	int inputID;
	int outputID;
	int method;
	bool debug;
    
public:
	MapToVector(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_name = object_cast<String>(parameters.get("METHOD"));
		if (method_name == "min_key") {
			method = METHOD_MIN_KEY;
		}
		else if (method_name == "max_key") {
			method = METHOD_MAX_KEY;
		}
		else if (method_name == "average") {
			method = METHOD_AVERAGE;
		}
		else if (method_name == "summation") {
			method = METHOD_SUMMATION;
		}
		else {
			cerr << "******* MapToVector: Input error" << endl;
			throw new NodeException(NULL, "MapToVector: METHOD should be \"min_key\" or \"max_key\" or \"average\" or \"summation\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;

		if (debug == true) {
			cout << "******* MapToVector debug point-Input check (count=" << count << ")" << endl;

			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Vector<float>))           cout << "Vector<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		if (input_map.empty() == true) {
			out[count] = ObjectRef(new Vector<float>);
		}
		else {
			int maxid = -1;
			int minid = 10000;    
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Vector<float>)) {
				const Vector<float> &vcFirst = object_cast<Vector<float> >(it->second);
				//ObjectRef orOutputTemp = ObjectRef(new Vector<float>(vcFirst.size()));
				//Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
				//out[count] = orOutputTemp;
				RCPtr<Vector<float> > vcFltOutput(new Vector<float>(vcFirst.size()));    
				out[count] = vcFltOutput;

				for (; it != input_map.end(); ++it) {
					const Vector<float> &vcFltInput = object_cast<Vector<float> >(it->second);
					if (vcFltInput.size() != vcFirst.size()) {
						break;
					}

					if (method == METHOD_MIN_KEY) {
						if (minid > it->first) {
							minid = it->first;
							for (int k = 0; k < vcFltInput.size(); k++) {
								//vcFltOutput[k] = vcFltInput[k];
								(*vcFltOutput)[k] = vcFltInput[k];
							}
						}
					}
					else if (method == METHOD_MAX_KEY) {
						if (maxid < it->first) {
							minid = it->first;
							for (int k = 0; k < vcFltInput.size(); k++) {
								//vcFltOutput[k] = vcFltInput[k];
								(*vcFltOutput)[k] = vcFltInput[k];
							}
						}
					}
					else { // method == METHOD_AVERAGE, METHOD_SUMMATION
						for (int k = 0; k < vcFltInput.size(); k++) {
							//vcFltOutput[k] += vcFltInput[k];
							(*vcFltOutput)[k] += vcFltInput[k];
						}
					}
				}

				if (it != input_map.end()) {
					cerr << "******* MapToVector: Input size of ObjectRef is not constant." << endl;
					throw(new NodeException(NULL, "Exception caught in calculate[MapToVector]", __FILE__, __LINE__));
				}
				else if (method == METHOD_AVERAGE) {
					//for (int k = 0; k < vcFltOutput.size(); k++) {
					for (int k = 0; k < vcFirst.size(); k++) {
						//vcFltOutput[k] /= input_map.size();
						(*vcFltOutput)[k] /= input_map.size();
					}
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<complex<float> >)) {
				const Vector<complex<float> > &vcFirst = object_cast<Vector<complex<float> > >(it->second);
				//ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(vcFirst.size()));
				//Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
				//out[count] = orOutputTemp;
				RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(vcFirst.size()));    
				out[count] = vcCmpOutput;

				for (; it != input_map.end(); ++it) {
					const Vector<complex<float> > &vcCmpInput = object_cast<Vector<complex<float> > >(it->second);
					if (vcCmpInput.size() != vcFirst.size()) {
						break;
					}

					if (method == METHOD_MIN_KEY) {
						if (minid > it->first) {
							minid = it->first;
							for (int k = 0; k < vcCmpInput.size(); k++) {
								//vcCmpOutput[k] = vcCmpInput[k];
								(*vcCmpOutput)[k] = vcCmpInput[k];
							}
						}
					}
					else if (method == METHOD_MAX_KEY) {
						if (maxid < it->first) {
							minid = it->first;
							for (int k = 0; k < vcCmpInput.size(); k++) {
								//vcCmpOutput[k] = vcCmpInput[k];
								(*vcCmpOutput)[k] = vcCmpInput[k];
							}
						}
					}
					else { // method == METHOD_AVERAGE, METHOD_SUMMATION
						for (int k = 0; k < vcCmpInput.size(); k++) {
							//vcCmpOutput[k] += vcCmpInput[k];
							(*vcCmpOutput)[k] += vcCmpInput[k];
						}
					}
				}

				if (it != input_map.end()) {
					cerr << "******* MapToVector: Input size of ObjectRef is not constant." << endl;
					throw(new NodeException(NULL, "Exception caught in calculate[MapToVector]", __FILE__, __LINE__));
				}
				else if (method == METHOD_AVERAGE) {
					//for (int k = 0; k < vcCmpOutput.size(); k++) {
					for (int k = 0; k < vcFirst.size(); k++) {
						//vcCmpOutput[k] /= input_map.size();
						(*vcCmpOutput)[k] /= input_map.size();
					}
				}
			}
			else {
				cerr << "******* MapToVector: Input type of ObjectRef is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapToVector]", __FILE__, __LINE__));
			}
		}

		if (debug == true) cout << "******* MapToVector debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(MapToVector);
};
