/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Matrix.h"

using namespace FD;
using namespace std;

class MatrixValueOverwrite;

DECLARE_NODE(MatrixValueOverwrite)
/*Node
 *
 * @name MatrixValueOverwrite
 * @category HARK:MISC
 * @description Replaces specific cells of the Matrix by a given value.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description The input Matrix. Matrix<int>, Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The Matrix with the overwritten value. Matrix<int>, Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name OVERWRITTEN_ROW_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the row where to start setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_ROW_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the row where to end setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_COL_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the column where to start setting the Matrix.
 *
 * @parameter_name OVERWRITTEN_COL_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index of the column where to end setting the Matrix.
 *
 * @parameter_name OVERWRITE_VALUE_REAL
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Real part of setting value.
 *
 * @parameter_name OVERWRITE_VALUE_IMAG
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Imaginary part of setting value.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

class MatrixValueOverwrite : public BufferedNode {
	int inputID;
	int outputID;
	int rwMin;
	int rwMax;
	int clMin;
	int clMax;
	float fReal;
	float fImag;	
	bool debug;

public:
	MatrixValueOverwrite(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		rwMin = dereference_cast<int>(parameters.get("OVERWRITTEN_ROW_MIN"));
		rwMax = dereference_cast<int>(parameters.get("OVERWRITTEN_ROW_MAX"));
		clMin = dereference_cast<int>(parameters.get("OVERWRITTEN_COL_MIN"));
		clMax = dereference_cast<int>(parameters.get("OVERWRITTEN_COL_MAX"));
		fReal = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_REAL"));
		fImag = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_IMAG"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void checkSize(int &vMin, int &vMax, int vSize) {
		if (vMin < 0) vMin = 0;
		else if (vMin >= vSize) vMin = vSize - 1;
		if (vMax < 0) vMax = 0;
		else if (vMax >= vSize) vMax = vSize - 1;
		if (vMin > vMax) {
			cerr << "******* MatrixValueOverwrite: Minimun index is larger than maximum index." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MatrixValueOverwrite]", __FILE__, __LINE__));
		}
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* MatrixValueOverwrite debug point-Input check (count=" << count << ")" << endl;
            cout << "Input type is: ";
			if (typeid(*input) == typeid(Matrix<int>))					cout << "Matrix<int>" << endl;
			else if (typeid(*input) == typeid(Matrix<float>))           cout << "Matrix<float>" << endl;
			else if (typeid(*input) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Matrix<int>)) {
			RCPtr<Matrix<int> > inputPtr = getInput(inputID, count);
			size_t rSize = inputPtr->nrows();
			size_t cSize = inputPtr->ncols();
			checkSize(rwMin, rwMax, rSize);
			checkSize(clMin, clMax, cSize);
			RCPtr<Matrix<int> > output(new Matrix<int>(rSize, cSize));
			/**output = *inputPtr;
			for (int i = rwMin; i <= rwMax; i++) {
				for (int j = clMin; j <= clMax; j++) {
					(*output)[i][j] = static_cast<int>(fReal);
				}
			}*/
			for (int i = 0; i < rSize; i++) {
				for (int j = 0; j < cSize; j++) {
					if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
						(*output)[i][j] = static_cast<int>(fReal);
					else
						(*output)[i][j] = (*inputPtr)[i][j];
				}
			}
		}
		else if (typeid(*input) == typeid(Matrix<float>)) {
			RCPtr<Matrix<float> > inputPtr = getInput(inputID, count);
			size_t rSize = inputPtr->nrows();
			size_t cSize = inputPtr->ncols();
			checkSize(rwMin, rwMax, rSize);
			checkSize(clMin, clMax, cSize);
			RCPtr<Matrix<float> > output(new Matrix<float>(rSize, cSize));
			/**output = *inputPtr;
			for (int i = rwMin; i <= rwMax; i++) {
				for (int j = clMin; j <= clMax; j++) {
					(*output)[i][j] = fReal;
				}
			}*/
			for (int i = 0; i < rSize; i++) {
				for (int j = 0; j < cSize; j++) {
					if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
						(*output)[i][j] = fReal;
					else
						(*output)[i][j] = (*inputPtr)[i][j];
				}
			}
			out[count] = output;
		}
		else if (typeid(*input) == typeid(Matrix<complex<float> >)) {
			RCPtr<Matrix<complex<float> > > inputPtr = getInput(inputID, count);
			size_t rSize = inputPtr->nrows();
			size_t cSize = inputPtr->ncols();
			checkSize(rwMin, rwMax, rSize);
			checkSize(clMin, clMax, cSize);
			RCPtr<Matrix<complex<float> > > output(new Matrix<complex<float> >(rSize, cSize));
			/**output = *inputPtr;
			for (int i = rwMin; i <= rwMax; i++) {
				for (int j = clMin; j <= clMax; j++) {
					(*output)[i][j] = complex<float>(fReal, fImag);
				}
			}*/
			for (int i = 0; i < rSize; i++) {
				for (int j = 0; j < cSize; j++) {
					if (i >= rwMin && i <= rwMax && j >= clMin && j <= clMax)
						(*output)[i][j] = complex<float>(fReal, fImag);
					else
						(*output)[i][j] = (*inputPtr)[i][j];
				}
			}
			out[count] = output;
		}
		else {
			cerr << "******* MatrixValueOverwrite: Input type is unidentified." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MatrixValueOverwrite]", __FILE__, __LINE__));
		}

        if (debug == true) cout << "******* MatrixValueOverwrite debug point-Finish" << endl;
	}

    NO_ORDER_NODE_SPEEDUP(MatrixValueOverwrite);
};
