/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Map.h"

using namespace std;
using namespace FD;

class ResizeMapVectorValues;

DECLARE_NODE(ResizeMapVectorValues)
/*Node
 *
 * @name ResizeMapVectorValues
 * @category HARK:MISC
 * @description Resize Vector value in Map<int,ObjectRef>. ObjectRef will be Vector<int>, Vector<float> or Vector<complex<float> >.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The input Map. Map<int,Vector<int> >, Map<int,Vector<float> > or Map<int,Vector<complex<float> > >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The Map with the resized value. Map<int,Vector<int> >, Map<int,Vector<float> > or Map<int,Vector<complex<float> > >.
 *
 * @parameter_name RESIZE_TYPE
 * @parameter_type string
 * @parameter_value RELATIVE
 * @parameter_list RELATIVE:ABSOLUTE
 * @parameter_description Resize type. Suppose current vector size is A. If RELATIVE, the vector size would be A+SIZE. If ABSOLUTE, the vector size would be SIZE. [default: RELATIVE] 
 *
 * @parameter_name SIZE
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Number of added elements. If negative, the elements will be deleted from the end. If positive, zero elements are added at the end. [default: 0]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int SIZE_RELATIVE = 0;
static const int SIZE_ABSOLUTE = 1;

class ResizeMapVectorValues : public BufferedNode {
	int inputID;
	int outputID;
	int resize_type;  
	int size;
	bool debug;

public:
	ResizeMapVectorValues(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string type_str = object_cast<String>(parameters.get("RESIZE_TYPE"));
		if (type_str == "RELATIVE")      resize_type = SIZE_RELATIVE;
		else if (type_str == "ABSOLUTE") resize_type = SIZE_ABSOLUTE;
		else {
			cerr << "******* ResizeMapVectorValues: Input error" << endl;
			throw new NodeException(NULL, "ResizeMapVectorValues: SIZE_TYPE should be \"RELATIVE\" or \"ABSOLUTE\".", __FILE__, __LINE__);
		}
		size = dereference_cast<int>(parameters.get("SIZE"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	int checkSize(int vSize) {
		int fsize = 0;
		if (resize_type == SIZE_RELATIVE)      fsize = vSize + size;
		else if (resize_type == SIZE_ABSOLUTE) fsize = size;
		if (fsize < 0) {
			cerr << "******* ResizeMapVectorValues: Resized size should not be negative." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[ResizeMapVectorValues]", __FILE__, __LINE__));
		}
		return fsize;
	}

	void calculate(int output_id, int count, Buffer &out) {
 		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;
     
		if (debug == true) {
			cout << "******* ResizeMapVectorValues debug point-Input check (count=" << count << ")" << endl;
 			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Vector<int>))			  cout << "Vector<int>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<float>))           cout << "Vector<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		Map<int, ObjectRef>& output = *new Map<int, ObjectRef>;
		out[count] = &output;

		if (input_map.empty() == false) {
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Vector<int>)) {
				for (; it != input_map.end(); ++it) {
					const Vector<int> &vcInput = object_cast<Vector<int> >(it->second);
					size_t vSize = vcInput.size();
					int fsize = checkSize(vSize);
					//RCPtr<Vector<int> > fbank(Vector<int>::alloc(fsize));
					RCPtr<Vector<int> > fbank(new Vector<int>(fsize));
					for (int i = 0; i < fsize; i++) {
						if (i >= vSize) break;
						(*fbank)[i] = vcInput[i];
					} 
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<float>)) {
				for (; it != input_map.end(); ++it) {
					const Vector<float> &vcInput = object_cast<Vector<float> >(it->second);
					size_t vSize = vcInput.size();
					int fsize = checkSize(vSize);
					//RCPtr<Vector<float> > fbank(Vector<float>::alloc(fsize));
					RCPtr<Vector<float> > fbank(new Vector<float>(fsize));
					for (int i = 0; i < fsize; i++) {
						if (i >= vSize) break;
						(*fbank)[i] = vcInput[i];
					} 
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Vector<complex<float> > &vcInput = object_cast<Vector<complex<float> > >(it->second);
					size_t vSize = vcInput.size();
					int fsize = checkSize(vSize);
					//RCPtr<Vector<complex<float> > > fbank(Vector<complex<float> >::alloc(fsize));
					RCPtr<Vector<complex<float> > > fbank(new Vector<complex<float> >(fsize));
					for (int i = 0; i < fsize; i++) {
						if (i >= vSize) break;
						(*fbank)[i] = vcInput[i];
					} 
					output[it->first] = fbank;
				}
			}
			else {
				cerr << "******* ResizeMapVectorValues: Input type is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[ResizeMapVectorValues]", __FILE__, __LINE__));
			}
		}		

        if (debug == true) cout << "******* ResizeMapVectorValues debug point-Finish" << endl;
	}

	IN_ORDER_NODE_SPEEDUP(ResizeMapVectorValues);
};
